#include <stdarg.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_internal.h"

struct buttondata {

	ggi_widget_size_t drawpos,drawsize;
	int doopt;
	int is_inside,will_activate;
};

static int button_getsize(ggi_widget_t self) {

	ggi_widget_t curchild;	

	self->min.x=self->min.y=0;
	self->opt.x=self->opt.y=0;

	curchild=self->getchild(self,0);
	if (curchild) {
		self->min=curchild->min;
		self->opt=curchild->opt;
	}
	self->min.x+=4;self->min.y+=4;
	self->opt.x+=8;self->opt.y+=8;
	return 0;
}

static int ggiWidgetButtonRedraw(ggi_widget_t self,int flush) {

	struct buttondata *mine=self->priv;
	ggi_widget_color_set *colset=ggiWidgetGetColorset(self);
	int rc;

	ggi_widget_size_t abspos;

	ggiWidgetDrawBorder(self,&mine->drawpos,&mine->drawsize);

	abspos.x=mine->drawpos.x+self->pos.x;
	abspos.y=mine->drawpos.y+self->pos.y;

	if (mine->will_activate) {
		ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background_dark));
	} else {
		ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background_light));
	}
	ggiDrawHLine(self->vis,abspos.x  ,abspos.y  ,mine->drawsize.x  );
	ggiDrawHLine(self->vis,abspos.x  ,abspos.y+1,mine->drawsize.x-1);
	ggiDrawVLine(self->vis,abspos.x  ,abspos.y+2,mine->drawsize.y-2);
	ggiDrawVLine(self->vis,abspos.x+1,abspos.y+2,mine->drawsize.y-3);

	if (!mine->will_activate) {
		ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background_dark));
	} else {
		ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background_light));
	}
	ggiDrawHLine(self->vis,abspos.x+1                 ,abspos.y+mine->drawsize.y-1,mine->drawsize.x-1);
	ggiDrawHLine(self->vis,abspos.x+2                 ,abspos.y+mine->drawsize.y-2,mine->drawsize.x-2);
	ggiDrawVLine(self->vis,abspos.x+mine->drawsize.x-1,abspos.y+1                 ,mine->drawsize.y-3);
	ggiDrawVLine(self->vis,abspos.x+mine->drawsize.x-2,abspos.y+2                 ,mine->drawsize.y-4);


	if (mine->doopt) {
		if (self->has_focus) {
			ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->text));
		} else {
			ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background));
		}
		ggiDrawBox(self->vis,abspos.x+2                 ,abspos.y+2                 ,mine->drawsize.x-4,2);
		ggiDrawBox(self->vis,abspos.x+2                 ,abspos.y+mine->drawsize.y-4,mine->drawsize.x-4,2);
		ggiDrawBox(self->vis,abspos.x+2                 ,abspos.y+4                 ,                 2,mine->drawsize.y-8);
		ggiDrawBox(self->vis,abspos.x+mine->drawsize.x-4,abspos.y+4                 ,                 2,mine->drawsize.y-8);
	}

	rc=ggiWidgetRedrawChildren(self);
	if (flush) {
		ggiFlushRegion(self->vis,self->pos.x,self->pos.y,self->size.x,self->size.y);
	}
	return rc;
}

static int ggiWidgetButtonVischanged(ggi_widget_t self) {

	struct buttondata *mine=self->priv;
	ggi_widget_t curchild;

	ggiWidgetGetPlacement(self,&mine->drawpos,&mine->drawsize,&mine->doopt,0,0);

	if ( NULL != (curchild=self->getchild(self,0)) ) {
		curchild->vis=self->vis;
		if (mine->doopt) {
			curchild->pos.x=self->pos.x+mine->drawpos.x+4;
			curchild->pos.y=self->pos.y+mine->drawpos.y+4;
			curchild->size.x=mine->drawsize.x-8;
			curchild->size.y=mine->drawsize.y-8;
		} else {
			curchild->pos.x=self->pos.x+mine->drawpos.x+2;
			curchild->pos.y=self->pos.y+mine->drawpos.y+2;
			curchild->size.x=mine->drawsize.x-4;;
			curchild->size.y=mine->drawsize.y-4;
		}
		curchild->vischanged(curchild);
	}

	return 0;
}

static int ggiWidgetButtonChildchanged(ggi_widget_t self,ggi_widget_t child) {

	button_getsize(self);	/* Recalc my size */
	
	if (self->parent) self->parent->childchanged(self->parent,self);
	else              self->vischanged(self);

	return 0;
}

int ggiWidgetButtonEvent(ggi_widget_t self,gii_event *ev,ggi_widget_inputstatus_t status) {

	struct buttondata *mine=self->priv;
	int is_inside;

	is_inside=(status->mouse.x>=self->pos.x+mine->drawpos.x &&
		   status->mouse.x< self->pos.x+mine->drawpos.x+mine->drawsize.x &&
		   status->mouse.y>=self->pos.y+mine->drawpos.y &&
		   status->mouse.y< self->pos.y+mine->drawpos.y+mine->drawsize.y );

	if (is_inside&&!mine->is_inside) {
		if (self->callback) {
			self->callback(GWT_CB_ENTER,self);
		}
	} else if (!is_inside&&mine->is_inside) {
		if (self->callback) {
			self->callback(GWT_CB_LEAVE,self);
		}
	};
	mine->is_inside=is_inside;

	switch(ev->any.type) {
		case evKeyPress:
		case evKeyRepeat:
			if (ev->key.sym==self->hotkey) {
				mine->will_activate=1;
				self->redraw(self,1);
				activate:
				if (self->statevar)
					*((int *)self->statevar)=1;
				if (self->callback) {
					self->callback(GWT_CB_ACTIVATE,self);
				}
				return 1;
			}
			if (self->has_focus&&ev->key.sym==' ') {
				mine->will_activate=1;
				self->redraw(self,1);
				goto activate;
			}
			break;
		case evKeyRelease:
			mine->will_activate=0;
			self->redraw(self,1);
			break;
		case evPtrButtonPress:
			if (is_inside) {
				ggiWidgetUnfocus(self);
				self->has_focus=1;
				if (ev->pbutton.button==GII_PBUTTON_PRIMARY) {
					mine->will_activate=1;
				}
				self->redraw(self,1);
			}
			break;
		case evPtrButtonRelease:
			if (is_inside) {
				if (ev->pbutton.button==GII_PBUTTON_PRIMARY&&
				    mine->will_activate) {
					mine->will_activate=0;
					self->redraw(self,1);
					goto activate;
				}
			}
			if (mine->will_activate) {
				mine->will_activate=0;
				self->redraw(self,1);
			}
			break;
	}
	return 0;
}

ggi_widget_t ggiWidgetCreateButton(va_list ap) {

	ggi_widget_t newid;
	ggi_widget_t parms;
	
	struct buttondata *newbut;

	newid=ggiWidgetAllocate();
	if (!newid) return NULL;

	if (NULL==(newbut=malloc(sizeof(*newbut)))) {
		free(newid);
		return NULL;
	}
	newbut->will_activate=newbut->is_inside=0;
	newid->priv=newbut;

	newid->redraw      =ggiWidgetButtonRedraw;
	newid->vischanged  =ggiWidgetButtonVischanged;
	newid->childchanged=ggiWidgetButtonChildchanged;

	newid->getchild   =ggiWidgetSingleChildGet;
	newid->linkchild  =ggiWidgetSingleChildLink;
	newid->unlinkchild=ggiWidgetSingleChildUnlink;

	newid->process_event=ggiWidgetButtonEvent;

	parms=va_arg(ap,ggi_widget_t);
	if (parms) newid->linkchild(newid,LAST_CHILD,parms);

	button_getsize(newid);

	return newid;
}
