#include <stdarg.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_internal.h"

struct containercarddata {

	ggi_widget_size_t drawpos,drawsize;
	int doopt;
};

static int containercard_getsize(ggi_widget_t self) {

	ggi_widget_t curchild;	
	int x;

	self->min.x=self->min.y=0;
	self->opt.x=self->opt.y=0;

	for(x=0;(curchild=self->getchild(self,x));x++) {
		if (self->min.x<curchild->min.x) {
			self->min.x=curchild->min.x;
		}
		if (self->min.y<curchild->min.y) {
			self->min.y=curchild->min.y;
		}
		if (self->opt.x<curchild->opt.x) {
			self->opt.x=curchild->opt.x;
		}
		if (self->opt.y<curchild->opt.y) {
			self->opt.y=curchild->opt.y;
		}
	}
	return 0;
}

static int ggiWidgetContainerCardRedraw(ggi_widget_t self,int flush) {

	struct containercarddata *mine=self->priv;
	int rc,childnum;
	ggi_widget_t child;

	ggiWidgetDrawBorder(self,&mine->drawpos,&mine->drawsize);

	childnum=0;
	if (self->statevar) 
		childnum=*(int *)self->statevar;

	child=self->getchild(self,childnum);
	rc=1;
	if (child) {
		rc=child->redraw(child,0);
	}
	if (flush) {
		ggiFlushRegion(self->vis,self->pos.x,self->pos.y,self->size.x,self->size.y);
	}
	return rc;
}

static int ggiWidgetContainerCardVischanged(ggi_widget_t self) {

	struct containercarddata *mine=self->priv;
	ggi_widget_t curchild;
	int x;

	ggiWidgetGetPlacement(self,&mine->drawpos,&mine->drawsize,&mine->doopt,0,0);

	for(x=0;(curchild=self->getchild(self,x));x++) {
		curchild->vis=self->vis;
		curchild->pos.x=self->pos.x+mine->drawpos.x;
		curchild->pos.y=self->pos.y+mine->drawpos.y;
		curchild->size.x=mine->drawsize.x;
		curchild->size.y=mine->drawsize.y;
		curchild->vischanged(curchild);
	}

	return 0;
}

static int ggiWidgetContainerCardChildchanged(ggi_widget_t self,ggi_widget_t child) {

	containercard_getsize(self);	/* Recalc my size */
	
	if (self->parent) self->parent->childchanged(self->parent,self);
	else              self->vischanged(self);

	return 0;
}

ggi_widget_t ggiWidgetCreateContainerCard(va_list ap) {

	ggi_widget_t newid;
	ggi_widget_t parms;
	
	struct containercarddata *newbut;

	newid=ggiWidgetAllocate();
	if (!newid) return NULL;

	if (NULL==(newbut=malloc(sizeof(*newbut)))) {
		free(newid);
		return NULL;
	}
	newid->priv=newbut;

	newid->redraw      =ggiWidgetContainerCardRedraw;
	newid->vischanged  =ggiWidgetContainerCardVischanged;
	newid->childchanged=ggiWidgetContainerCardChildchanged;

	newid->getchild   =ggiWidgetChildlistGet;
	newid->linkchild  =ggiWidgetChildlistLink;
	newid->unlinkchild=ggiWidgetChildlistUnlink;

	while( (parms=va_arg(ap,ggi_widget_t)) ) {
		newid->linkchild(newid,LAST_CHILD,parms);
	}

	containercard_getsize(newid);

	return newid;
}
