#include <stdarg.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_internal.h"

struct containerstackdata {

	ggi_widget_size_t drawpos,drawsize;
	int doopt;
	int padx,pady;
};

static int containerstack_getsize(ggi_widget_t self) {

	struct containerstackdata *mine=self->priv;
	ggi_widget_t curchild;	
	int x;

	self->min.x=self->min.y=0;
	self->opt.x=self->opt.y=0;

	for(x=0;(curchild=self->getchild(self,x));x++) {
		self->min.y+=curchild->min.y+2*mine->pady;
		if (self->min.x<curchild->min.x) {
			self->min.x=curchild->min.x;
		}
		self->opt.y+=curchild->opt.y+2*mine->pady;
		if (self->opt.x<curchild->opt.x) {
			self->opt.x=curchild->opt.x;
		}
	}
	self->min.x+=2*mine->padx;
	self->opt.x+=2*mine->padx;
	return 0;
}

static int ggiWidgetContainerStackRedraw(ggi_widget_t self,int flush) {

	struct containerstackdata *mine=self->priv;
	int rc;

	ggiWidgetDrawBorder(self,&mine->drawpos,&mine->drawsize);
	rc=ggiWidgetRedrawChildren(self);
	if (flush) {
		ggiFlushRegion(self->vis,self->pos.x,self->pos.y,self->size.x,self->size.y);
	}
	return rc;
}

static int ggiWidgetContainerStackVischanged(ggi_widget_t self) {

	struct containerstackdata *mine=self->priv;
	ggi_widget_t curchild;
	int mysizeremain,x,numchilds,dy,add;

	ggiWidgetGetPlacement(self,&mine->drawpos,&mine->drawsize,&mine->doopt,0,0);

	mysizeremain   =mine->drawsize.y - 
			(mine->doopt ? self->opt.y : self->min.y);

	for(numchilds=0;self->getchild(self,numchilds);numchilds++);

	for(x=dy=0;x<numchilds;x++) {
		curchild=self->getchild(self,x);
		curchild->vis=self->vis;
		curchild->pos.x=self->pos.x+mine->drawpos.x;
		curchild->pos.y=self->pos.y+mine->drawpos.y+dy;
		curchild->size.x=mine->drawsize.x;
		curchild->size.y=mine->doopt ? curchild->opt.y : curchild->min.y;
		add=mysizeremain/(numchilds-x);
		curchild->size.y+=add+2*mine->pady;
		mysizeremain-=add;
		dy+=curchild->size.y;
		curchild->vischanged(curchild);
	}

	return 0;
}

static int ggiWidgetContainerStackChildchanged(ggi_widget_t self,ggi_widget_t child) {

	containerstack_getsize(self);	/* Recalc my size */
	
	if (self->parent) self->parent->childchanged(self->parent,self);
	else              self->vischanged(self);

	return 0;
}

ggi_widget_t ggiWidgetCreateContainerStack(va_list ap) {

	ggi_widget_t newid;
	ggi_widget_t parms;
	
	struct containerstackdata *newstack;

	newid=ggiWidgetAllocate();
	if (!newid) return NULL;

	if (NULL==(newstack=malloc(sizeof(*newstack)))) {
		free(newid);
		return NULL;
	}
	newstack->padx=va_arg(ap,int);
	newstack->pady=va_arg(ap,int);
	newid->priv=newstack;

	newid->redraw      =ggiWidgetContainerStackRedraw;
	newid->vischanged  =ggiWidgetContainerStackVischanged;
	newid->childchanged=ggiWidgetContainerStackChildchanged;

	newid->getchild   =ggiWidgetChildlistGet;
	newid->linkchild  =ggiWidgetChildlistLink;
	newid->unlinkchild=ggiWidgetChildlistUnlink;

	while( (parms=va_arg(ap,ggi_widget_t)) ) {
		newid->linkchild(newid,LAST_CHILD,parms);
	}

	containerstack_getsize(newid);

	return newid;
}
