#include <stdarg.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_internal.h"

struct framedata {

	ggi_widget_size_t drawpos,drawsize;
	int doopt;
};

static int frame_getsize(ggi_widget_t self) {

	ggi_widget_t curchild;	

	self->min.x=self->min.y=0;
	self->opt.x=self->opt.y=0;

	curchild=self->getchild(self,0);
	if (curchild) {
		self->min=curchild->min;
		self->opt=curchild->opt;
	}
	self->min.x+=4;self->min.y+=4;
	self->opt.x+=8;self->opt.y+=8;
	return 0;
}

static int ggiWidgetFrameRedraw(ggi_widget_t self,int flush) {

	struct framedata *mine=self->priv;
	ggi_widget_size_t abspos;
	ggi_widget_color_set *colset=ggiWidgetGetColorset(self); 
	int rc;

	ggiWidgetDrawBorder(self,&mine->drawpos,&mine->drawsize);

	abspos.x=mine->drawpos.x+self->pos.x;
	abspos.y=mine->drawpos.y+self->pos.y;

	ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background_light));
	ggiDrawHLine(self->vis,abspos.x  ,abspos.y  ,mine->drawsize.x  );
	ggiDrawHLine(self->vis,abspos.x  ,abspos.y+1,mine->drawsize.x-1);
	ggiDrawVLine(self->vis,abspos.x  ,abspos.y+2,mine->drawsize.y-2);
	ggiDrawVLine(self->vis,abspos.x+1,abspos.y+2,mine->drawsize.y-3);

	ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background_dark));
	ggiDrawHLine(self->vis,abspos.x+1                 ,abspos.y+mine->drawsize.y-1,mine->drawsize.x-1);
	ggiDrawHLine(self->vis,abspos.x+2                 ,abspos.y+mine->drawsize.y-2,mine->drawsize.x-2);
	ggiDrawVLine(self->vis,abspos.x+mine->drawsize.x-1,abspos.y+1                 ,mine->drawsize.y-3);
	ggiDrawVLine(self->vis,abspos.x+mine->drawsize.x-2,abspos.y+2                 ,mine->drawsize.y-4);


	if (mine->doopt) {
		ggiSetGCForeground(self->vis,ggiMapColor(self->vis,&colset->background));
		ggiDrawBox(self->vis,abspos.x+2                 ,abspos.y+2                 ,mine->drawsize.x-4,2);
		ggiDrawBox(self->vis,abspos.x+2                 ,abspos.y+mine->drawsize.y-4,mine->drawsize.x-4,2);
		ggiDrawBox(self->vis,abspos.x+2                 ,abspos.y+4                 ,                 2,mine->drawsize.y-8);
		ggiDrawBox(self->vis,abspos.x+mine->drawsize.x-4,abspos.y+4                 ,                 2,mine->drawsize.y-8);
	}

	rc=ggiWidgetRedrawChildren(self);
	if (flush) {
		ggiFlushRegion(self->vis,self->pos.x,self->pos.y,self->size.x,self->size.y);
	}
	return rc;
}

static int ggiWidgetFrameVischanged(ggi_widget_t self) {

	struct framedata *mine=self->priv;
	ggi_widget_t curchild;

	ggiWidgetGetPlacement(self,&mine->drawpos,&mine->drawsize,&mine->doopt,0,0);

	if ( NULL != (curchild=self->getchild(self,0)) ) {
		curchild->vis=self->vis;
		if (mine->doopt) {
			curchild->pos.x=self->pos.x+mine->drawpos.x+4;
			curchild->pos.y=self->pos.y+mine->drawpos.y+4;
			curchild->size.x=mine->drawsize.x-8;
			curchild->size.y=mine->drawsize.y-8;
		} else {
			curchild->pos.x=self->pos.x+mine->drawpos.x+2;
			curchild->pos.y=self->pos.y+mine->drawpos.y+2;
			curchild->size.x=mine->drawsize.x-4;;
			curchild->size.y=mine->drawsize.y-4;
		}
		curchild->vischanged(curchild);
	}

	return 0;
}

static int ggiWidgetFrameChildchanged(ggi_widget_t self,ggi_widget_t child) {

	frame_getsize(self);	/* Recalc my size */
	
	if (self->parent) self->parent->childchanged(self->parent,self);
	else              self->vischanged(self);

	return 0;
}

ggi_widget_t ggiWidgetCreateFrame(va_list ap) {

	ggi_widget_t newid;
	ggi_widget_t parms;
	
	struct framedata *newbut;

	newid=ggiWidgetAllocate();
	if (!newid) return NULL;

	if (NULL==(newbut=malloc(sizeof(*newbut)))) {
		free(newid);
		return NULL;
	}
	newid->priv=newbut;

	newid->redraw      =ggiWidgetFrameRedraw;
	newid->vischanged  =ggiWidgetFrameVischanged;
	newid->childchanged=ggiWidgetFrameChildchanged;

	newid->getchild   =ggiWidgetSingleChildGet;
	newid->linkchild  =ggiWidgetSingleChildLink;
	newid->unlinkchild=ggiWidgetSingleChildUnlink;

	parms=va_arg(ap,ggi_widget_t);
	if (parms) newid->linkchild(newid,LAST_CHILD,parms);

	frame_getsize(newid);

	return newid;
}
