#include <stdio.h>
#include <ggi/ggi.h>
#include <math.h>
#include <string.h>
#include "ggi_widget.h"
#include "ggi_widget_highlevel.h"
#include "ggi_widget_bitmap.h"

int quit=0;

static struct voxrendata {
	FILE *comm;
	double step;
	int autopilot_on,autopilot_mode;
	int pivotmode,render_mode,light_mode;
	double angle,jump,light;
	int exactedge,smoothedge,smoothsurface,hidedata;
	int sympal,invpal,showcut;
	int size,scale;
	unsigned char commandbuffer[256];
} vrdata = {
	NULL,
	1.0,
	0,0,
	0,0,0,
	100,0,1000,
	0,0,0,0,
	0,0,0,
	1,0,
	""
};

static void say_step(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_ACTIVATE) return;
	fprintf(vrdata.comm,"%s %f\n",(char *)widget->callbackpriv,vrdata.step);
}

static void say_command(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_ACTIVATE) return;
	fprintf(vrdata.comm,"%s\n",vrdata.commandbuffer);
}

static void say_fix(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_ACTIVATE) return;
	fprintf(vrdata.comm,"%s\n",(char *)widget->callbackpriv);
}

static void say_double(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"%s %f\n",(char *)widget->callbackpriv,*((double *)widget->statevar));
}

static void say_int(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"%s %d\n",(char *)widget->callbackpriv,*((int *)widget->statevar));
}

static void say_flags(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"FLAGS %d\n",(vrdata.sympal)|(vrdata.invpal<<1)|(vrdata.showcut<<2)|(vrdata.hidedata<<3));
}

static void say_smooth(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"ALIAS %d\n",(vrdata.smoothedge)|(vrdata.smoothsurface<<1));
}

static void say_zoom(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"%s %f\n",(char *)widget->callbackpriv,2.0*atan(*((double *)widget->statevar)/180.0*M_PI));
}

static void say_autopilot(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"AUTO %d\n",vrdata.autopilot_on ? vrdata.autopilot_mode+1 : 0);
}

static const char *pivot_modes[]={
	"Camera   ",
	"VolCenter",
	"PicCenter",
	"Cutplane ",
	NULL
};

static void say_pivot(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"TMODE %s\n",pivot_modes[vrdata.pivotmode]);
}

static const char *render_modes[]={
	"RGB",
	"QRG",
	"XYZ",
	"XYT",
	"SMN",
	"ZSH",
	"KES",
	"KE",
	NULL
};

static void say_rmode(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"RENDERMODE %s\n",render_modes[vrdata.render_mode]);
}

static const char *light_modes[]={
	"CENTER",
	"TOPC",
	"TLC",
	"TRC",
	"BRC",
	"BLC",
	"PARC",
	"PARALLEL",
	NULL
};

static void say_lmode(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"LIGHTMODE %s\n",light_modes[vrdata.light_mode]);
}

static const char *size_modes[]={
	" 96x 96",
	"256x256",
	"320x240",
	"512x512",
	"640x480",
	NULL
};

static void say_size(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"SIZEX %3.3s\n",size_modes[vrdata.size]);
	fprintf(vrdata.comm,"SIZEY %3.3s\n",size_modes[vrdata.size]+4);
}

static const char *scale_modes[]={
	"1",
	"2",
	"3",
	"4",
	"5",
	NULL
};

static void say_scale(ggiWidgetCallbackType cbt,ggi_widget_t widget) {

	if (cbt!=GWT_CB_STATECHANGE) return;
	fprintf(vrdata.comm,"SCALE %s\n",scale_modes[vrdata.scale]);
}

ggi_widget_bitmap_t leftbmp,rightbmp,upbmp,downbmp,aroundbmp;

static ggi_widget_t setup_slide(void) {

	ggi_widget_t grid,up,down,left,right,lab;

	grid=ggiWidgetCreateWidget(GWT_CONTAINER_GRID,4,3);

#define MKSLIDEBUTTON(name,bmp,cmd,x,y) \
	name=ggiWidgetMakeBitmapbutton(bmp);	\
	name->callback=say_step;		\
	name->callbackpriv=cmd;			\
	grid->linkchild(grid,LAST_CHILD,name,x,y);

	MKSLIDEBUTTON(up   ,upbmp   ,"UP"   ,1,0)
	MKSLIDEBUTTON(down ,downbmp ,"DOWN" ,1,2)
	MKSLIDEBUTTON(left ,leftbmp ,"LEFT" ,0,1)
	MKSLIDEBUTTON(right,rightbmp,"RIGHT",2,1)

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Slide");
	grid->linkchild(grid,LAST_CHILD,lab,1,1);

	MKSLIDEBUTTON(up   ,upbmp   ,"FWD"  ,3,0)
	MKSLIDEBUTTON(down ,downbmp ,"REV"  ,3,2)

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Go");
	grid->linkchild(grid,LAST_CHILD,lab,3,1);
	
	return grid;
}

static ggi_widget_t setup_turn(void) {

	ggi_widget_t grid,up,down,left,right,lab;

	grid=ggiWidgetCreateWidget(GWT_CONTAINER_GRID,3,3);

	MKSLIDEBUTTON(up   ,upbmp   ,"TUP"   ,1,0)
	MKSLIDEBUTTON(down ,downbmp ,"TDOWN" ,1,2)
	MKSLIDEBUTTON(left ,leftbmp ,"TLEFT" ,0,1)
	MKSLIDEBUTTON(right,rightbmp,"TRIGHT",2,1)

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Turn");
	grid->linkchild(grid,LAST_CHILD,lab,1,1);
	
	return grid;
}

static ggi_widget_t setup_roll(void) {

	ggi_widget_t grid,left,right,lab;

	grid=ggiWidgetCreateWidget(GWT_CONTAINER_GRID,3,1);

	MKSLIDEBUTTON(left ,leftbmp ,"RLEFT" ,0,0)
	MKSLIDEBUTTON(right,rightbmp,"RRIGHT",2,0)
#undef MKSLIDEBUTTON

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Roll");
	grid->linkchild(grid,LAST_CHILD,lab,1,0);
	
	return grid;
}

static ggi_widget_t setup_turn90(void) {

	ggi_widget_t grid,up,down,left,right,lab,around;

	grid=ggiWidgetCreateWidget(GWT_CONTAINER_GRID,3,3);

#define MKFIXBUTTON(name,bmp,cmd,x,y) \
	name=ggiWidgetMakeBitmapbutton(bmp);	\
	name->callback=say_fix;			\
	name->callbackpriv=cmd;			\
	grid->linkchild(grid,LAST_CHILD,name,x,y);

	MKFIXBUTTON(up   ,upbmp   ,"TUP 90"   ,1,0)
	MKFIXBUTTON(down ,downbmp ,"TDOWN 90" ,1,2)
	MKFIXBUTTON(left ,leftbmp ,"TLEFT 90" ,0,1)
	MKFIXBUTTON(right,rightbmp,"TRIGHT 90",2,1)
	MKFIXBUTTON(around,aroundbmp,"UTURN"  ,2,2)

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Turn\n90");
	grid->linkchild(grid,LAST_CHILD,lab,1,1);

#undef MKFIXBUTTON

	return grid;
}

static struct abspos{
	char x[6],y[6],z[6];
} abspos = {
"\0\0\0\0\0\0","\0\0\0\0\0\0","\0\0\0\0\0\0"
};

static ggi_widget_t setup_abspos(void) {

	ggi_widget_t clinex,cliney,clinez,midstack,lab,txt,send;

	lab =ggiWidgetCreateWidget(GWT_LABEL,"X:");
	txt =ggiWidgetCreateWidget(GWT_TEXT,abspos.x,5,5);
	clinex=ggiWidgetCreateWidget(GWT_CONTAINER_LINE,2,2,lab,txt,NULL);

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Y:");
	txt =ggiWidgetCreateWidget(GWT_TEXT,abspos.y,5,5);
	cliney=ggiWidgetCreateWidget(GWT_CONTAINER_LINE,2,2,lab,txt,NULL);

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Z:");
	txt =ggiWidgetCreateWidget(GWT_TEXT,abspos.z,5,5);
	clinez=ggiWidgetCreateWidget(GWT_CONTAINER_LINE,2,2,lab,txt,NULL);

	send=ggiWidgetMakeTextbutton("Send");

	midstack=ggiWidgetCreateWidget(GWT_CONTAINER_STACK,0,1,clinex,cliney,clinez,send,NULL);
	
	return midstack;
}

static ggi_widget_t setup_autopilot(void) {

	ggi_widget_t stack,lab,cbox,type;
	
	const char *modes[]={
		"Normal",
		"Gravity",
		"Mouse",
		"Drop",
		NULL
	};

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Autopilot");
	cbox=ggiWidgetCreateWidget(GWT_CHECKBOX,lab,NULL);
	cbox->callback=say_autopilot;
	cbox->statevar=&vrdata.autopilot_on;
	type=ggiWidgetCreateWidget(GWT_OPTIONCYCLER,modes);
	type->callback=say_autopilot;
	type->statevar=&vrdata.autopilot_mode;

	stack=ggiWidgetCreateWidget(GWT_CONTAINER_STACK,0,1,cbox,type,NULL);
	
	return stack;
}

static ggi_widget_t setup_pivot(void) {

	ggi_widget_t stack,lab,type;
	

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Pivot Center");
	type=ggiWidgetCreateWidget(GWT_OPTIONCYCLER,pivot_modes);
	type->callback=say_pivot;
	type->statevar=&vrdata.pivotmode;

	stack=ggiWidgetCreateWidget(GWT_CONTAINER_STACK,0,1,lab,type,NULL);
	
	return stack;
}


static ggi_widget_t setup_move(void) {

	ggi_widget_t slide,turn,turn90,abspos,apilot,pivot,step,roll;
	ggi_widget_t grid;

	slide =setup_slide();
	turn  =setup_turn();
	turn90=setup_turn90();
	abspos=setup_abspos();
	apilot=setup_autopilot();
	pivot =setup_pivot();
	roll  =setup_roll();

	step=ggiWidgetCreateWidget(GWT_DIAL,"Step",25,30.0,330.0,0.0,10.0,4,1,0.1,1.0,4,1);
	step->statevar=&vrdata.step;

	grid=ggiWidgetCreateWidget(GWT_CONTAINER_GRID,4,2);
	grid->linkchild(grid,LAST_CHILD,slide ,0,0);
	grid->linkchild(grid,LAST_CHILD,turn  ,1,0);
	grid->linkchild(grid,LAST_CHILD,turn90,2,0);
	grid->linkchild(grid,LAST_CHILD,abspos,3,0);
	grid->linkchild(grid,LAST_CHILD,step  ,0,1);
	grid->linkchild(grid,LAST_CHILD,roll  ,1,1);
	grid->linkchild(grid,LAST_CHILD,pivot ,2,1);
	grid->linkchild(grid,LAST_CHILD,apilot,3,1);
	grid->gravity=GWT_GRAV_STRETCH_X;

	return grid;
}


static ggi_widget_t setup_options(void) {

	ggi_widget_t lab,hdda,exed,smed,smsu,stack;

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Hide Data");
	hdda=ggiWidgetCreateWidget(GWT_CHECKBOX,lab,NULL);
	hdda->statevar=&vrdata.hidedata;
	hdda->callback=say_flags;
	hdda->gravity=GWT_GRAV_WEST;
	lab =ggiWidgetCreateWidget(GWT_LABEL,"Exact Edges");
	exed=ggiWidgetCreateWidget(GWT_CHECKBOX,lab,NULL);
	exed->statevar=&vrdata.exactedge;
	exed->callback=say_int;
	exed->callbackpriv="FINE";
	exed->gravity=GWT_GRAV_WEST;
	lab =ggiWidgetCreateWidget(GWT_LABEL,"Smooth Edges");
	smed=ggiWidgetCreateWidget(GWT_CHECKBOX,lab,NULL);
	smed->statevar=&vrdata.smoothedge;
	smed->callback=say_smooth;
	smed->gravity=GWT_GRAV_WEST;
	lab =ggiWidgetCreateWidget(GWT_LABEL,"Smooth Surfaces");
	smsu=ggiWidgetCreateWidget(GWT_CHECKBOX,lab,NULL);
	smsu->statevar=&vrdata.smoothsurface;
	smsu->callback=say_smooth;
	smsu->gravity=GWT_GRAV_WEST;
	lab =ggiWidgetCreateWidget(GWT_LABEL,"Render Options:");

	stack=ggiWidgetCreateWidget(GWT_CONTAINER_STACK,0,1,lab,hdda,exed,smed,smsu,NULL);

	return stack;
}

static ggi_widget_t setup_rendermode(void) {

	ggi_widget_t cline,lab,rmode;

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Render-\nMode:");
	rmode=ggiWidgetCreateWidget(GWT_OPTIONCYCLER,render_modes);
	rmode->callback=say_rmode;
	rmode->statevar=&vrdata.render_mode;

	cline=ggiWidgetCreateWidget(GWT_CONTAINER_LINE,1,0,lab,rmode,NULL);
	
	return cline;
}

static ggi_widget_t setup_lightmode(void) {

	ggi_widget_t lmode;

	lmode=ggiWidgetCreateWidget(GWT_OPTIONCYCLER,light_modes);
	lmode->callback=say_lmode;
	lmode->statevar=&vrdata.light_mode;
	return lmode;
}

static ggi_widget_t setup_camparms(void) {

	ggi_widget_t grid,angle,light,cp,opts,lab,showcut,rmode,lmode;
	
	angle=ggiWidgetCreateWidget(GWT_DIAL,"Angle",40,30.0,330.0,5.0, 175.0,16,0, 5.0,15.0,4,1);
	angle->statevar=&vrdata.angle;
	angle->callback=say_zoom;
	angle->callbackpriv="ZOOM";
	cp   =ggiWidgetCreateWidget(GWT_DIAL,"Cut"  ,40,30.0,330.0,0.0, 999.0, 9,0, 1.0,10.0,5,1);
	cp->statevar=&vrdata.jump;
	cp->callback=say_double;
	cp->callbackpriv="JUMP";
	light=ggiWidgetCreateWidget(GWT_DIAL,"Light",40,30.0,330.0,0.0,9999.0, 9,0,10.0,50.0,6,1);
	light->statevar=&vrdata.light;
	light->callback=say_double;
	light->callbackpriv="LIGHT";
	opts=setup_options();


	lab =ggiWidgetCreateWidget(GWT_LABEL,"Show\nCutplane");
	showcut=ggiWidgetCreateWidget(GWT_CHECKBOX,lab,NULL);
	showcut->statevar=&vrdata.showcut;
	showcut->callback=say_flags;

	lmode=setup_lightmode();

	rmode=setup_rendermode();


	grid=ggiWidgetCreateWidget(GWT_CONTAINER_GRID,4,2);
	grid->linkchild(grid,LAST_CHILD,angle  ,0,0);
	grid->linkchild(grid,LAST_CHILD,cp     ,1,0);
	grid->linkchild(grid,LAST_CHILD,light  ,2,0);
	grid->linkchild(grid,LAST_CHILD,opts   ,3,0);
//	grid->linkchild(grid,LAST_CHILD,step   ,0,1);
	grid->linkchild(grid,LAST_CHILD,showcut,1,1);
	grid->linkchild(grid,LAST_CHILD,lmode  ,2,1);
	grid->linkchild(grid,LAST_CHILD,rmode  ,3,1);
	grid->gravity=GWT_GRAV_STRETCH_X;

	return grid;
}

static ggi_widget_t setup_size(void) {

	ggi_widget_t lab,cline,lab2,size,scale;

	lab =ggiWidgetCreateWidget(GWT_LABEL,"Size:");
	size=ggiWidgetCreateWidget(GWT_OPTIONCYCLER,size_modes);
	size->callback=say_size;
	size->statevar=&vrdata.size;
	lab2=ggiWidgetCreateWidget(GWT_LABEL,"x");
	scale=ggiWidgetCreateWidget(GWT_OPTIONCYCLER,scale_modes);
	scale->callback=say_scale;
	scale->statevar=&vrdata.scale;

	cline=ggiWidgetCreateWidget(GWT_CONTAINER_LINE,4,0,lab,size,lab2,scale,NULL);
	return cline;
}

static ggi_widget_t setup_command(void) {

	ggi_widget_t lab,txt,cline;
	
	lab=ggiWidgetCreateWidget(GWT_LABEL,"Command");
	memset(vrdata.commandbuffer,0,sizeof(vrdata.commandbuffer));
	txt   =ggiWidgetCreateWidget(GWT_TEXT,vrdata.commandbuffer,40,sizeof(vrdata.commandbuffer)-1);
	txt->callback=say_command;
	        
	cline=ggiWidgetCreateWidget(GWT_CONTAINER_LINE,4,0,lab,txt,NULL);

	return cline;
}

static ggi_widget_t setup_quit(void) {

	ggi_widget_t button;
	
	button=ggiWidgetMakeTextbutton("QUIT");
	button->statevar=&quit;

	return button;
}

static ggi_widget_t setup_all(void) {

	ggi_widget_t move,camp,size,cmd,quit,stack;

	move=setup_move();
	camp=setup_camparms();
	size=setup_size();
	cmd =setup_command();
	quit=setup_quit();
	stack=ggiWidgetCreateWidget(GWT_CONTAINER_STACK,0,0,move,camp,size,cmd,quit,NULL);
	stack->gravity=GWT_GRAV_STRETCH;

	return stack;
}

int main(int argc,char *argv[]) {

	ggi_visual_t vis;
	ggi_widget_t top;
	
	ggiInit();
	vis=ggiOpen(NULL);
	
	if (!vis) {
		fprintf(stderr,"Huh ?\n");
		ggiExit();
		return 1;
	}
	
	ggiSetFlags(vis,GGIFLAG_ASYNC);
	ggiSetSimpleMode(vis,512,512,1,GT_AUTO);

	vrdata.comm=stdout;
	setlinebuf(vrdata.comm);

	leftbmp =ggiWidgetBitmapLoadPBM("left.pbm");
	rightbmp=ggiWidgetBitmapLoadPBM("right.pbm");
	upbmp   =ggiWidgetBitmapLoadPBM("up.pbm");
	downbmp =ggiWidgetBitmapLoadPBM("down.pbm");
	aroundbmp=ggiWidgetBitmapLoadPBM("around.pbm");

	top=setup_all();
	ggiWidgetAttach(top,vis,0,0,512,512);
	top->redraw(top,1);

	while(!quit) {
		ggi_event ev;
		ggiEventRead(vis,&ev,emAll);	
		ggiWidgetProcessEvent(vis,&ev);
	}

	ggiWidgetDetach(top,vis);
	top->destroy(top);

	ggiClose(vis);
	ggiExit();
	return 0;
}
